import logging
import sys
import time
from typing import Any, Optional, Union
import grpc

from . import types
from .internal import channel_provider
from .shared.admin_helpers import BaseClient

logger = logging.getLogger(__name__)


class Client(BaseClient):
    """
    Aerospike Vector Search Admin Client

    This client is designed to conduct Aerospike Vector Search administrative operation such as creating indexes, querying index information, and dropping indexes.

    :param seeds: Defines the AVS nodes to which you want AVS to connect. AVS iterates through the seed nodes. After connecting to a node, AVS discovers all of the nodes in the cluster.
    :type seeds: Union[types.HostPort, tuple[types.HostPort, ...]]

    :param listener_name: An external (NATed) address and port combination that differs from the actual address and port where AVS is listening. Clients can access AVS on a node using the advertised listener address and port. Defaults to None.
    :type listener_name: Optional[str]

    :param is_loadbalancer: If true, the first seed address will be treated as a load balancer node. Defaults to False.
    :type is_loadbalancer: Optional[bool]

    :param service_config_path: Path to the service configuration file. Defaults to None.
    :type service_config_path: Optional[str]

    :param username: Username for Role-Based Access. Defaults to None.
    :type username: Optional[str]

    :param password: Password for Role-Based Access. Defaults to None.
    :type password: Optional[str]

    :param root_certificate: The PEM-encoded root certificates as a byte string. Defaults to None.
    :type root_certificate: Optional[list[bytes], bytes]

    :param certificate_chain: The PEM-encoded certificate chain as a byte string. Defaults to None.
    :type certificate_chain: Optional[bytes]

    :param private_key: The PEM-encoded private key as a byte string. Defaults to None.
    :type private_key: Optional[bytes]

    :raises AVSClientError: Raised when no seed host is provided.

    """

    def __init__(
        self,
        *,
        seeds: Union[types.HostPort, tuple[types.HostPort, ...]],
        listener_name: Optional[str] = None,
        is_loadbalancer: Optional[bool] = False,
        username: Optional[str] = None,
        password: Optional[str] = None,
        root_certificate: Optional[Union[list[str], str]] = None,
        certificate_chain: Optional[str] = None,
        private_key: Optional[str] = None,
        service_config_path: Optional[str] = None,
    ) -> None:
        seeds = self._prepare_seeds(seeds)

        self._channel_provider = channel_provider.ChannelProvider(
            seeds,
            listener_name,
            is_loadbalancer,
            username,
            password,
            root_certificate,
            certificate_chain,
            private_key,
            service_config_path,
        )

    def index_create(
        self,
        *,
        namespace: str,
        name: str,
        vector_field: str,
        dimensions: int,
        vector_distance_metric: Optional[types.VectorDistanceMetric] = (
            types.VectorDistanceMetric.SQUARED_EUCLIDEAN
        ),
        sets: Optional[str] = None,
        index_params: Optional[types.HnswParams] = None,
        index_labels: Optional[dict[str, str]] = None,
        index_storage: Optional[types.IndexStorage] = None,
        timeout: Optional[int] = None,
    ) -> None:
        """
        Create an index.

        :param namespace: The namespace for the index.
        :type namespace: str

        :param name: The name of the index.
        :type name: str

        :param vector_field: The name of the field containing vector data.
        :type vector_field: str

        :param dimensions: The number of dimensions in the vector data.
        :type dimensions: int

        :param vector_distance_metric:
            The distance metric used to compare when performing a vector search.
            Defaults to :class:`VectorDistanceMetric.SQUARED_EUCLIDEAN`.
        :type vector_distance_metric: Optional[types.VectorDistanceMetric]

        :param sets: The set used for the index. Defaults to None.
        :type sets: Optional[str]

        :param index_params: (Optional[types.HnswParams], optional): Parameters used for tuning
            vector search. Defaults to None. If index_params is None, then the default values
            specified for :class:`types.HnswParams` will be used.
        :type index_params: Optional[types.HnswParams]

        :param index_labels: Meta data associated with the index. Defaults to None.
        :type index_labels: Optional[dict[str, str]]

        :param index_storage: Namespace and set where index overhead (non-vector data) is stored.
        :type index_storage: Optional[types.IndexStorage]

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to create the index.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        Note:
            This method creates an index with the specified parameters and waits for the index creation to complete.
            It waits for up to 100,000 seconds for the index creation to complete.
        """

        (index_stub, index_create_request, kwargs) = self._prepare_index_create(
            namespace,
            name,
            vector_field,
            dimensions,
            vector_distance_metric,
            sets,
            index_params,
            index_labels,
            index_storage,
            timeout,
            logger,
        )

        try:
            index_stub.Create(
                index_create_request,
                credentials=self._channel_provider.get_token(),
                **kwargs,
            )
        except grpc.RpcError as e:
            logger.error("Failed to create index with error: %s", e)
            raise types.AVSServerError(rpc_error=e)
        try:
            self._wait_for_index_creation(
                namespace=namespace, name=name, timeout=100_000
            )
        except grpc.RpcError as e:
            logger.error("Failed waiting for creation with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

    def index_drop(
        self, *, namespace: str, name: str, timeout: Optional[int] = None
    ) -> None:
        """
        Drop an index.

        :param namespace: The namespace of the index.
        :type name: str

        :param name: The name of the index.
        :type name: str

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to drop the index.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        Note:
            This method drops an index with the specified parameters and waits for the index deletion to complete.
            It waits for up to 100,000 seconds for the index deletion to complete.
        """

        (index_stub, index_drop_request, kwargs) = self._prepare_index_drop(
            namespace, name, timeout, logger
        )

        try:
            index_stub.Drop(
                index_drop_request,
                credentials=self._channel_provider.get_token(),
                **kwargs,
            )
        except grpc.RpcError as e:
            logger.error("Failed to drop index with error: %s", e)
            raise types.AVSServerError(rpc_error=e)
        try:
            self._wait_for_index_deletion(
                namespace=namespace, name=name, timeout=100_000
            )
        except grpc.RpcError as e:
            logger.error("Failed waiting for deletion with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

    def index_list(self, timeout: Optional[int] = None) -> list[dict]:
        """
        List all indices.

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        Returns: list[dict]: A list of indices.

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to list the index.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.
        """

        (index_stub, index_list_request, kwargs) = self._prepare_index_list(
            timeout, logger
        )

        try:
            response = index_stub.List(
                index_list_request,
                credentials=self._channel_provider.get_token(),
                **kwargs,
            )
        except grpc.RpcError as e:
            logger.error("Failed to list indexes with error: %s", e)
            raise types.AVSServerError(rpc_error=e)
        return self._respond_index_list(response)

    def index_get(
        self, *, namespace: str, name: str, timeout: Optional[int] = None
    ) -> dict[str, Union[int, str]]:
        """
        Retrieve the information related with an index.

        :param namespace: The namespace of the index.
        :type namespace: str

        :param name: The name of the index.
        :type name: str

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        Returns: dict[str, Union[int, str]: Information about an index.

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to get the index.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        """

        (index_stub, index_get_request, kwargs) = self._prepare_index_get(
            namespace, name, timeout, logger
        )

        try:
            response = index_stub.Get(
                index_get_request,
                credentials=self._channel_provider.get_token(),
                **kwargs,
            )
        except grpc.RpcError as e:
            logger.error("Failed to get index with error: %s", e)
            raise types.AVSServerError(rpc_error=e)
        return self._respond_index_get(response)

    def index_get_status(
        self, *, namespace: str, name: str, timeout: Optional[int] = None
    ) -> int:
        """
        Retrieve the number of records queued to be merged into an index.

        :param namespace: The namespace of the index.
        :type name: str

        :param name: The name of the index.
        :type name: str

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        Returns: int: Records queued to be merged into an index.

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to get the index status.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        Note:
            This method retrieves the status of the specified index. If index_get_status is called the vector client puts some records into Aerospike Vector Search,
            the records may not immediately begin to merge into the index. To wait for all records to be merged into an index, use vector_client.wait_for_index_completion.

            Warning: This API is subject to change.
        """
        (index_stub, index_get_status_request, kwargs) = self._prepare_index_get_status(
            namespace, name, timeout, logger
        )

        try:
            response = index_stub.GetStatus(
                index_get_status_request,
                credentials=self._channel_provider.get_token(),
                **kwargs,
            )
        except grpc.RpcError as e:
            logger.error("Failed to get index status with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

        return self._respond_index_get_status(response)

    def add_user(
        self,
        *,
        username: str,
        password: str,
        roles: list[str],
        timeout: Optional[int] = None,
    ) -> None:
        """
        Add role-based access AVS User to the AVS Server.

        :param username: Username for the new user.
        :type username: str

        :param password: Password for the new user.
        :type password: str

        :param roles: Roles for the new user.
        :type password: list[str]

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int


        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to add a user.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        """
        (user_admin_stub, add_user_request, kwargs) = self._prepare_add_user(
            username, password, roles, timeout, logger
        )

        try:
            user_admin_stub.AddUser(
                add_user_request,
                credentials=self._channel_provider.get_token(),
                **kwargs,
            )
        except grpc.RpcError as e:
            logger.error("Failed to add user with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

    def update_credentials(
        self, *, username: str, password: str, timeout: Optional[int] = None
    ) -> None:
        """
        Update AVS User credentials.

        :param username: Username of the user to update.
        :type username: str

        :param password: New password for the user.
        :type password: str

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int


        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to update a users credentials.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        """
        (user_admin_stub, update_credentials_request, kwargs) = (
            self._prepare_update_credentials(username, password, timeout, logger)
        )

        try:
            user_admin_stub.UpdateCredentials(
                update_credentials_request,
                credentials=self._channel_provider.get_token(),
                **kwargs,
            )
        except grpc.RpcError as e:
            logger.error("Failed to update credentials with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

    def drop_user(self, *, username: str, timeout: Optional[int] = None) -> None:
        """
        Drops AVS User from the AVS Server.

        :param username: Username of the user to drop.
        :type username: str

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int


        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to drop a user
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        """
        (user_admin_stub, drop_user_request, kwargs) = self._prepare_drop_user(
            username, timeout, logger
        )

        try:
            user_admin_stub.DropUser(
                drop_user_request,
                credentials=self._channel_provider.get_token(),
                **kwargs,
            )
        except grpc.RpcError as e:
            logger.error("Failed to drop user with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

    def get_user(self, *, username: str, timeout: Optional[int] = None) -> types.User:
        """
        Retrieves AVS User information from the AVS Server.

        :param username: Username of the user to be retrieved.
        :type username: str

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        return: types.User: AVS User

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to get a user.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        """
        (user_admin_stub, get_user_request, kwargs) = self._prepare_get_user(
            username, timeout, logger
        )

        try:
            response = user_admin_stub.GetUser(
                get_user_request,
                credentials=self._channel_provider.get_token(),
                **kwargs,
            )
        except grpc.RpcError as e:
            logger.error("Failed to get user with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

        return self._respond_get_user(response)

    def list_users(self, timeout: Optional[int] = None) -> list[types.User]:
        """
        List all users existing on the AVS Server.

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        return: list[types.User]: list of AVS Users

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to list users.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        """
        (user_admin_stub, list_users_request, kwargs) = self._prepare_list_users(
            timeout, logger
        )

        try:
            response = user_admin_stub.ListUsers(
                list_users_request,
                credentials=self._channel_provider.get_token(),
                **kwargs,
            )
        except grpc.RpcError as e:
            logger.error("Failed to list user with error: %s", e)
            raise types.AVSServerError(rpc_error=e)
        return self._respond_list_users(response)

    def grant_roles(
        self, *, username: str, roles: list[str], timeout: Optional[int] = None
    ) -> None:
        """
        Grant roles to existing AVS Users.

        :param username: Username of the user which will receive the roles.
        :type username: str

        :param roles: Roles the specified user will recieved.
        :type roles: list[str]

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to grant roles.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        """
        (user_admin_stub, grant_roles_request, kwargs) = self._prepare_grant_roles(
            username, roles, timeout, logger
        )

        try:
            user_admin_stub.GrantRoles(
                grant_roles_request,
                credentials=self._channel_provider.get_token(),
                **kwargs,
            )
        except grpc.RpcError as e:
            logger.error("Failed to grant roles with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

    def revoke_roles(
        self, *, username: str, roles: list[str], timeout: Optional[int] = None
    ) -> None:
        """
        Revoke roles from existing AVS Users.

        :param username: Username of the user undergoing role removal.
        :type username: str

        :param roles: Roles to be revoked.
        :type roles: list[str]

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to revoke roles.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        """
        (user_admin_stub, revoke_roles_request, kwargs) = self._prepare_revoke_roles(
            username, roles, timeout, logger
        )

        try:
            user_admin_stub.RevokeRoles(
                revoke_roles_request,
                credentials=self._channel_provider.get_token(),
                **kwargs,
            )
        except grpc.RpcError as e:
            logger.error("Failed to revoke roles with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

    def list_roles(self, timeout: Optional[int] = None) -> list[dict]:
        """
        List roles available on the AVS server.

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        returns: list[str]: Roles available in the AVS Server.

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to list roles.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.
        """
        (user_admin_stub, list_roles_request, kwargs) = self._prepare_list_roles(
            timeout, logger
        )

        try:
            response = user_admin_stub.ListRoles(
                list_roles_request,
                credentials=self._channel_provider.get_token(),
                **kwargs,
            )
        except grpc.RpcError as e:
            logger.error("Failed to list roles with error: %s", e)
            raise types.AVSServerError(rpc_error=e)
        return self._respond_list_roles(response)

    def _wait_for_index_creation(
        self,
        *,
        namespace: str,
        name: str,
        timeout: Optional[int] = sys.maxsize,
        wait_interval: Optional[int] = 0.1,
    ) -> None:
        """
        Wait for the index to be created.
        """

        (index_stub, wait_interval, start_time, _, _, index_creation_request) = (
            self._prepare_wait_for_index_waiting(namespace, name, wait_interval)
        )
        while True:
            self._check_timeout(start_time, timeout)
            try:
                index_stub.GetStatus(
                    index_creation_request,
                    credentials=self._channel_provider.get_token(),
                )
                logger.debug("Index created succesfully")
                # Index has been created
                return
            except grpc.RpcError as e:
                if e.code() == grpc.StatusCode.NOT_FOUND:

                    # Wait for some more time.
                    time.sleep(wait_interval)
                else:
                    logger.error("Failed waiting for index creation with error: %s", e)
                    raise types.AVSServerError(rpc_error=e)

    def _wait_for_index_deletion(
        self,
        *,
        namespace: str,
        name: str,
        timeout: Optional[int] = sys.maxsize,
        wait_interval: Optional[int] = 0.1,
    ) -> None:
        """
        Wait for the index to be deleted.
        """

        # Wait interval between polling
        (index_stub, wait_interval, start_time, _, _, index_deletion_request) = (
            self._prepare_wait_for_index_waiting(namespace, name, wait_interval)
        )

        while True:
            self._check_timeout(start_time, timeout)

            try:
                index_stub.GetStatus(
                    index_deletion_request,
                    credentials=self._channel_provider.get_token(),
                )
                # Wait for some more time.
                time.sleep(wait_interval)
            except grpc.RpcError as e:
                if e.code() == grpc.StatusCode.NOT_FOUND:
                    logger.debug("Index deleted succesfully")
                    # Index has been created
                    return
                else:
                    logger.error("Failed waiting for index deletion with error: %s", e)
                    raise types.AVSServerError(rpc_error=e)

    def close(self):
        """
        Close the Aerospike Vector Search Admin Client.

        This method closes gRPC channels connected to Aerospike Vector Search.

        Note:
            This method should be called when the VectorDbAdminClient is no longer needed to release resources.
        """
        self._channel_provider.close()

    def __enter__(self):
        """
        Enter a context manager for the admin client.

        Returns:
            VectorDbAdminlient: Aerospike Vector Search Admin Client instance.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit a context manager for the admin client.
        """
        self.close()



import re
import time
import logging
import threading
from typing import Optional, Union

import google.protobuf.empty_pb2
import grpc

from .. import types
from ..shared.proto_generated import vector_db_pb2
from ..shared.proto_generated import vector_db_pb2_grpc
from ..shared import base_channel_provider

empty = google.protobuf.empty_pb2.Empty()

logger = logging.getLogger(__name__)

TEND_INTERVAL = 1

class ChannelProvider(base_channel_provider.BaseChannelProvider):
    """Proximus Channel Provider"""

    def __init__(
        self,
        seeds: tuple[types.HostPort, ...],
        listener_name: Optional[str] = None,
        is_loadbalancer: Optional[bool] = False,
        username: Optional[str] = None,
        password: Optional[str] = None,
        root_certificate: Optional[Union[list[str], str]] = None,
        certificate_chain: Optional[str] = None,
        private_key: Optional[str] = None,
        service_config_path: Optional[str] = None,
    ) -> None:
        super().__init__(
            seeds,
            listener_name,
            is_loadbalancer,
            username,
            password,
            root_certificate,
            certificate_chain,
            private_key,
            service_config_path,
        )
        # When set, client has concluded cluster tending
        self._tend_ended = threading.Event()

        # When locked, new task is being assigned to _auth_task
        self._auth_tending_lock: threading.Lock = threading.Lock()

        self._auth_timer = None

        # initializes authentication tending
        self._tend_token()

        # verfies server is minimally compatible with client
        self._check_server_version()

        # initializes cluster tending
        self._tend_cluster()





    def _tend_cluster(self):
        try:
            (channels, end_tend_cluster) = (
                self.init_tend_cluster()
            )

            if end_tend_cluster:
                self._tend_ended.set()
                return

            (cluster_info_stubs, new_cluster_ids) = self._gather_new_cluster_ids_and_cluster_info_stubs(channels)

            update_endpoints_stubs = self._gather_stubs_for_endpoint_updating(new_cluster_ids, cluster_info_stubs)

            cluster_endpoints_list = self._gather_temp_endpoints(new_cluster_ids, update_endpoints_stubs)

            temp_endpoints = self._assign_temporary_endpoints(cluster_endpoints_list)

            if update_endpoints_stubs:

                self._add_new_channels_from_temp_endpoints(temp_endpoints)

                self._close_old_channels_from_node_channels(temp_endpoints)


            threading.Timer(TEND_INTERVAL, self._tend_cluster).start()

        except Exception as e:
            logger.error("Tending failed at unindentified location: %s", e)
            raise e
    
    def _call_get_cluster_id(self, stub):
        try:
            return stub.GetClusterId(empty, credentials=self._token,)
        except Exception as e:
            logger.debug(
                "While tending, failed to get cluster id with error: " + str(e)
            )

    def _call_get_cluster_endpoints(self, stub):
        try:
            return (
                stub.GetClusterEndpoints(
                    vector_db_pb2.ClusterNodeEndpointsRequest(
                        listenerName=self.listener_name
                    ),
                    credentials=self._token,
                )
            ).endpoints
        except Exception as e:
            logger.debug(
                "While tending, failed to get cluster endpoints with error: "
                + str(e)
            )

    def _call_close_on_channel(self, channel_endpoints):
        try:
            channel_endpoints.channel.close()
        except Exception as e:
            logger.debug(
                "While tending, failed to close GRPC channel: "
                + str(e)
            )

    def _tend_token(self):

        if not self._token:
            return

        self._update_token_and_ttl()

        with self._auth_tending_lock:
            self._auth_timer = threading.Timer((self._ttl * self._ttl_threshold), self._tend_token)
            self._auth_timer.start()



    def _update_token_and_ttl(
        self,
    ) -> None:

        (auth_stub, auth_request) = self._prepare_authenticate(
            self._credentials, logger
        )

        try:
            response = auth_stub.Authenticate(auth_request)
        except grpc.RpcError as e:
            logger.error("Failed to refresh authentication token with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

        self._respond_authenticate(response.token)

    def _check_server_version(self):
        stub = vector_db_pb2_grpc.AboutServiceStub(self.get_channel())
        about_request = vector_db_pb2.AboutRequest()

        try:
            response = stub.Get(
                about_request, credentials=self._token
            )
            self.current_server_version = response.version
        except grpc.RpcError as e:
            logger.debug(
                "Failed to retrieve server version: "
                + str(e)
            )
            raise AVSServerError(rpc_error=e)  
        self.verify_compatibile_server()



    def _create_channel(self, host: str, port: int) -> grpc.Channel:
        host = re.sub(r"%.*", "", host)

        if self.service_config_json:
            options = []
            options.append(("grpc.service_config", self.service_config_json))
        else:
            options = None

        if self._root_certificate:

            ssl_credentials = grpc.ssl_channel_credentials(
                root_certificates=self._root_certificate,
                certificate_chain=self._certificate_chain,
                private_key=self._private_key,
            )

            return grpc.secure_channel(
                f"{host}:{port}", ssl_credentials, options=options
            )

        else:
            return grpc.insecure_channel(f"{host}:{port}", options=options)

    def close(self):
        self._closed = True
        self._tend_ended.wait()

        for channel in self._seedChannels:
            channel.close()

        for k, channelEndpoints in self._node_channels.items():
            if channelEndpoints.channel:
                channelEndpoints.channel.close()

        with self._auth_tending_lock:
            if self._auth_timer != None:
                self._auth_timer.cancel()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: vector-db.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0fvector-db.proto\x12\x10\x61\x65rospike.vector\x1a\x1bgoogle/protobuf/empty.proto\"\x0e\n\x0c\x41\x62outRequest\" \n\rAboutResponse\x12\x0f\n\x07version\x18\x01 \x01(\t\"\x17\n\tClusterId\x12\n\n\x02id\x18\x01 \x01(\x04\"\x14\n\x06NodeId\x12\n\n\x02id\x18\x01 \x01(\x04\">\n\x0eServerEndpoint\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\r\x12\r\n\x05isTls\x18\x03 \x01(\x08\"I\n\x12ServerEndpointList\x12\x33\n\tendpoints\x18\x01 \x03(\x0b\x32 .aerospike.vector.ServerEndpoint\"\xb8\x01\n\x14\x43lusterNodeEndpoints\x12H\n\tendpoints\x18\x01 \x03(\x0b\x32\x35.aerospike.vector.ClusterNodeEndpoints.EndpointsEntry\x1aV\n\x0e\x45ndpointsEntry\x12\x0b\n\x03key\x18\x01 \x01(\x04\x12\x33\n\x05value\x18\x02 \x01(\x0b\x32$.aerospike.vector.ServerEndpointList:\x02\x38\x01\"I\n\x1b\x43lusterNodeEndpointsRequest\x12\x19\n\x0clistenerName\x18\x01 \x01(\tH\x00\x88\x01\x01\x42\x0f\n\r_listenerName\"\x81\x01\n\x0f\x43lusteringState\x12\x13\n\x0bisInCluster\x18\x01 \x01(\x08\x12.\n\tclusterId\x18\x02 \x01(\x0b\x32\x1b.aerospike.vector.ClusterId\x12)\n\x07members\x18\x03 \x03(\x0b\x32\x18.aerospike.vector.NodeId2X\n\x0c\x41\x62outService\x12H\n\x03Get\x12\x1e.aerospike.vector.AboutRequest\x1a\x1f.aerospike.vector.AboutResponse\"\x00\x32\xdf\x02\n\x12\x43lusterInfoService\x12?\n\tGetNodeId\x12\x16.google.protobuf.Empty\x1a\x18.aerospike.vector.NodeId\"\x00\x12\x45\n\x0cGetClusterId\x12\x16.google.protobuf.Empty\x1a\x1b.aerospike.vector.ClusterId\"\x00\x12Q\n\x12GetClusteringState\x12\x16.google.protobuf.Empty\x1a!.aerospike.vector.ClusteringState\"\x00\x12n\n\x13GetClusterEndpoints\x12-.aerospike.vector.ClusterNodeEndpointsRequest\x1a&.aerospike.vector.ClusterNodeEndpoints\"\x00\x42\x43\n!com.aerospike.vector.client.protoP\x01Z\x1c\x61\x65rospike.com/vector/protos/b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'vector_db_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n!com.aerospike.vector.client.protoP\001Z\034aerospike.com/vector/protos/'
  _globals['_CLUSTERNODEENDPOINTS_ENDPOINTSENTRY']._loaded_options = None
  _globals['_CLUSTERNODEENDPOINTS_ENDPOINTSENTRY']._serialized_options = b'8\001'
  _globals['_ABOUTREQUEST']._serialized_start=66
  _globals['_ABOUTREQUEST']._serialized_end=80
  _globals['_ABOUTRESPONSE']._serialized_start=82
  _globals['_ABOUTRESPONSE']._serialized_end=114
  _globals['_CLUSTERID']._serialized_start=116
  _globals['_CLUSTERID']._serialized_end=139
  _globals['_NODEID']._serialized_start=141
  _globals['_NODEID']._serialized_end=161
  _globals['_SERVERENDPOINT']._serialized_start=163
  _globals['_SERVERENDPOINT']._serialized_end=225
  _globals['_SERVERENDPOINTLIST']._serialized_start=227
  _globals['_SERVERENDPOINTLIST']._serialized_end=300
  _globals['_CLUSTERNODEENDPOINTS']._serialized_start=303
  _globals['_CLUSTERNODEENDPOINTS']._serialized_end=487
  _globals['_CLUSTERNODEENDPOINTS_ENDPOINTSENTRY']._serialized_start=401
  _globals['_CLUSTERNODEENDPOINTS_ENDPOINTSENTRY']._serialized_end=487
  _globals['_CLUSTERNODEENDPOINTSREQUEST']._serialized_start=489
  _globals['_CLUSTERNODEENDPOINTSREQUEST']._serialized_end=562
  _globals['_CLUSTERINGSTATE']._serialized_start=565
  _globals['_CLUSTERINGSTATE']._serialized_end=694
  _globals['_ABOUTSERVICE']._serialized_start=696
  _globals['_ABOUTSERVICE']._serialized_end=784
  _globals['_CLUSTERINFOSERVICE']._serialized_start=787
  _globals['_CLUSTERINFOSERVICE']._serialized_end=1138
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import types_pb2 as types__pb2
from . import user_admin_pb2 as user__admin__pb2

GRPC_GENERATED_VERSION = '1.64.1'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.65.0'
SCHEDULED_RELEASE_DATE = 'June 25, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in user_admin_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class UserAdminServiceStub(object):
    """User admin service
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.AddUser = channel.unary_unary(
                '/aerospike.vector.UserAdminService/AddUser',
                request_serializer=user__admin__pb2.AddUserRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.UpdateCredentials = channel.unary_unary(
                '/aerospike.vector.UserAdminService/UpdateCredentials',
                request_serializer=user__admin__pb2.UpdateCredentialsRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.DropUser = channel.unary_unary(
                '/aerospike.vector.UserAdminService/DropUser',
                request_serializer=user__admin__pb2.DropUserRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.GetUser = channel.unary_unary(
                '/aerospike.vector.UserAdminService/GetUser',
                request_serializer=user__admin__pb2.GetUserRequest.SerializeToString,
                response_deserializer=types__pb2.User.FromString,
                _registered_method=True)
        self.ListUsers = channel.unary_unary(
                '/aerospike.vector.UserAdminService/ListUsers',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=user__admin__pb2.ListUsersResponse.FromString,
                _registered_method=True)
        self.GrantRoles = channel.unary_unary(
                '/aerospike.vector.UserAdminService/GrantRoles',
                request_serializer=user__admin__pb2.GrantRolesRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.RevokeRoles = channel.unary_unary(
                '/aerospike.vector.UserAdminService/RevokeRoles',
                request_serializer=user__admin__pb2.RevokeRolesRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.ListRoles = channel.unary_unary(
                '/aerospike.vector.UserAdminService/ListRoles',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=user__admin__pb2.ListRolesResponse.FromString,
                _registered_method=True)


class UserAdminServiceServicer(object):
    """User admin service
    """

    def AddUser(self, request, context):
        """Add a new user.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateCredentials(self, request, context):
        """Update user credentials.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DropUser(self, request, context):
        """Drop a user.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetUser(self, request, context):
        """Get details for a user.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListUsers(self, request, context):
        """List users.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GrantRoles(self, request, context):
        """Grant roles to a user.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RevokeRoles(self, request, context):
        """Revoke roles from a user.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListRoles(self, request, context):
        """List roles.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_UserAdminServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'AddUser': grpc.unary_unary_rpc_method_handler(
                    servicer.AddUser,
                    request_deserializer=user__admin__pb2.AddUserRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'UpdateCredentials': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateCredentials,
                    request_deserializer=user__admin__pb2.UpdateCredentialsRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'DropUser': grpc.unary_unary_rpc_method_handler(
                    servicer.DropUser,
                    request_deserializer=user__admin__pb2.DropUserRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'GetUser': grpc.unary_unary_rpc_method_handler(
                    servicer.GetUser,
                    request_deserializer=user__admin__pb2.GetUserRequest.FromString,
                    response_serializer=types__pb2.User.SerializeToString,
            ),
            'ListUsers': grpc.unary_unary_rpc_method_handler(
                    servicer.ListUsers,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=user__admin__pb2.ListUsersResponse.SerializeToString,
            ),
            'GrantRoles': grpc.unary_unary_rpc_method_handler(
                    servicer.GrantRoles,
                    request_deserializer=user__admin__pb2.GrantRolesRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'RevokeRoles': grpc.unary_unary_rpc_method_handler(
                    servicer.RevokeRoles,
                    request_deserializer=user__admin__pb2.RevokeRolesRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'ListRoles': grpc.unary_unary_rpc_method_handler(
                    servicer.ListRoles,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=user__admin__pb2.ListRolesResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'aerospike.vector.UserAdminService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('aerospike.vector.UserAdminService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class UserAdminService(object):
    """User admin service
    """

    @staticmethod
    def AddUser(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.UserAdminService/AddUser',
            user__admin__pb2.AddUserRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def UpdateCredentials(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.UserAdminService/UpdateCredentials',
            user__admin__pb2.UpdateCredentialsRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DropUser(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.UserAdminService/DropUser',
            user__admin__pb2.DropUserRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetUser(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.UserAdminService/GetUser',
            user__admin__pb2.GetUserRequest.SerializeToString,
            types__pb2.User.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListUsers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.UserAdminService/ListUsers',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            user__admin__pb2.ListUsersResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GrantRoles(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.UserAdminService/GrantRoles',
            user__admin__pb2.GrantRolesRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RevokeRoles(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.UserAdminService/RevokeRoles',
            user__admin__pb2.RevokeRolesRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListRoles(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.UserAdminService/ListRoles',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            user__admin__pb2.ListRolesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



import logging
import random
import time

from typing import Optional, Union

import json
import jwt
import re
import grpc


from .. import types
from . import helpers

from .proto_generated import vector_db_pb2, auth_pb2
from .proto_generated import auth_pb2_grpc
from .proto_generated import vector_db_pb2_grpc

logger = logging.getLogger(__name__)


class ChannelAndEndpoints(object):
    def __init__(
        self,
        channel: Union[grpc.Channel, grpc.aio.Channel],
        endpoints: vector_db_pb2.ServerEndpointList,
    ) -> None:
        self.channel = channel
        self.endpoints = endpoints


class BaseChannelProvider(object):
    """AVS Channel Provider"""

    def __init__(
        self,
        seeds: tuple[types.HostPort, ...],
        listener_name: Optional[str] = None,
        is_loadbalancer: Optional[bool] = False,
        username: Optional[str] = None,
        password: Optional[str] = None,
        root_certificate: Optional[str] = None,
        certificate_chain: Optional[str] = None,
        private_key: Optional[str] = None,
        service_config_path: Optional[str] = None,
    ) -> None:
        self.seeds: tuple[types.HostPort, ...] = seeds
        self.listener_name: Optional[str] = listener_name
        self._is_loadbalancer: Optional[bool] = is_loadbalancer

        if service_config_path:
            with open(service_config_path, "rb") as f:
                self.service_config_json = f.read()
        else:
            self.service_config_json = None

        self._credentials = helpers._get_credentials(username, password)
        if self._credentials:
            self._token = True
        else:
            self._token = None
        self._root_certificate = root_certificate
        self._certificate_chain = certificate_chain
        self._private_key = private_key
        self._ttl = 0
        self._ttl_start = 0
        self._ttl_threshold = 0.9
        # dict of Node Number and ChannelAndEndponts object
        self._node_channels: dict[int, ChannelAndEndpoints] = {}
        self._seedChannels: Union[list[grpc.Channel], list[grpc.Channel.aio]] = [
            self._create_channel_from_host_port(seed) for seed in self.seeds
        ]
        self._closed: bool = False
        self._cluster_id: int = 0
        self.current_server_version = ""
        self.minimum_required_version = "0.9.0"
        self.client_server_compatible = False

    def get_token(self) -> grpc.access_token_call_credentials:
        return self._token

    def _prepare_about(self) -> grpc.Channel:
        stub = vector_db_pb2_grpc.AboutServiceStub(self.get_channel())
        about_request = vector_db_pb2.AboutRequest()
        return (stub, about_request)

    def get_channel(self) -> Union[grpc.aio.Channel, grpc.Channel]:
        if not self._is_loadbalancer:
            discovered_channels: list[ChannelAndEndpoints] = list(
                self._node_channels.values()
            )
            if len(discovered_channels) <= 0:
                return self._seedChannels[0]

            # Return a random channel.
            channel = random.choice(discovered_channels).channel
            if channel:
                return channel

        return self._seedChannels[0]

    def _create_channel_from_host_port(
        self, host: types.HostPort
    ) -> Union[grpc.aio.Channel, grpc.Channel]:
        return self._create_channel(host.host, host.port)

    def _create_channel_from_server_endpoint_list(
        self, endpoints: vector_db_pb2.ServerEndpointList
    ) -> Union[grpc.aio.Channel, grpc.Channel]:
        # TODO: Create channel with all endpoints
        for endpoint in endpoints.endpoints:
            if ":" in endpoint.address:
                # TODO: Ignoring IPv6 for now. Needs fix
                continue
            try:
                return self._create_channel(endpoint.address, endpoint.port)
            except Exception as e:
                logger.debug("failure creating channel: " + str(e))

    def add_new_channel_to_node_channels(self, node, newEndpoints):

        # We have discovered a new node
        new_channel = self._create_channel_from_server_endpoint_list(newEndpoints)
        self._node_channels[node] = ChannelAndEndpoints(new_channel, newEndpoints)

    def init_tend_cluster(self) -> None:

        end_tend_cluster = False
        if self._is_loadbalancer or self._closed:
            # Skip tend if we are behind a load-balancer
            end_tend_cluster = True

        channels = self._seedChannels + [
            x.channel for x in self._node_channels.values()
        ]

        return (channels, end_tend_cluster)

    def check_cluster_id(self, new_cluster_id) -> None:
        if new_cluster_id == self._cluster_id:
            return False

        self._cluster_id = new_cluster_id

        return True

    def update_temp_endpoints(self, endpoints, temp_endpoints):
        if len(endpoints) > len(temp_endpoints):
            return endpoints
        else:
            return temp_endpoints

    def check_for_new_endpoints(self, node, newEndpoints):

        channel_endpoints = self._node_channels.get(node)

        add_new_channel = True

        if channel_endpoints:
            # We have this node. Check if the endpoints changed.
            if channel_endpoints.endpoints == newEndpoints:
                # Nothing to be done for this node
                add_new_channel = False

        return (channel_endpoints, add_new_channel)

    def _get_ttl(self, payload):
        return payload["exp"] - payload["iat"]

    def _prepare_authenticate(self, credentials, logger):
        logger.debug("Refreshing auth token")
        auth_stub = self._get_auth_stub()

        auth_request = self._get_authenticate_request(self._credentials)

        return (auth_stub, auth_request)

    def _get_auth_stub(self):
        return auth_pb2_grpc.AuthServiceStub(self.get_channel())

    def _get_authenticate_request(self, credentials):
        return auth_pb2.AuthRequest(credentials=credentials)

    def _respond_authenticate(self, token):
        payload = jwt.decode(
            token, "", algorithms=["RS256"], options={"verify_signature": False}
        )
        self._ttl = self._get_ttl(payload)
        self._ttl_start = payload["iat"]

        self._token = grpc.access_token_call_credentials(token)

    def verify_compatibile_server(self) -> bool:
        def parse_version(v: str):
            return tuple(str(part) if part.isdigit() else part for part in v.split("."))
        if parse_version(self.current_server_version) < parse_version(self.minimum_required_version):
            self._tend_ended.set()
            raise types.AVSClientError(
                message="This AVS Client version is only compatbile with AVS Servers above the following version number: "
                + self.minimum_required_version
            )
        else:
            self.client_server_compatible = True

    def _gather_new_cluster_ids_and_cluster_info_stubs(self, channels):

        stubs = []
        responses = []

        for channel in channels:

            stub = vector_db_pb2_grpc.ClusterInfoServiceStub(channel)
            stubs.append(stub)

            response = self._call_get_cluster_id(stub)
            responses.append(response)

        return (stubs, responses)

    def _gather_stubs_for_endpoint_updating(self, new_cluster_ids, cluster_info_stubs):
        update_endpoints_stubs = []
        for index, value in enumerate(new_cluster_ids):

            if self.check_cluster_id(value.id):
                update_endpoints_stub = cluster_info_stubs[index]

                update_endpoints_stubs.append(update_endpoints_stub)
        return update_endpoints_stubs

    def _gather_temp_endpoints(self, new_cluster_ids, update_endpoints_stubs):

        responses = []
        for stub in update_endpoints_stubs:
            response = self._call_get_cluster_endpoints(stub)


            responses.append(response)
        return responses

    def _assign_temporary_endpoints(self, cluster_endpoints_list):
        # TODO: Worry about thread safety
        temp_endpoints: dict[int, vector_db_pb2.ServerEndpointList] = {}
        for endpoints in cluster_endpoints_list:
            temp_endpoints = self.update_temp_endpoints(
                endpoints, temp_endpoints
            )
        return temp_endpoints


    def _add_new_channels_from_temp_endpoints(self, temp_endpoints):
        responses = []


        for node, newEndpoints in temp_endpoints.items():

            # Compare node channel result
            (channel_endpoints, add_new_channel) = self.check_for_new_endpoints(
                node, newEndpoints
            )


            if add_new_channel:
                if channel_endpoints:
                    response = self._call_close_on_channel(channel_endpoints)
                    responses.append(response)
                
                self.add_new_channel_to_node_channels(node, newEndpoints)

        return responses

    def _close_old_channels_from_node_channels(self, temp_endpoints):
        responses = []


        for node, channel_endpoints in list(self._node_channels.items()):
            if not temp_endpoints.get(node):
                # TODO: Wait for all calls to drain
                response = self._call_close_on_channel(channel_endpoints)
                responses.append(response)


                del self._node_channels[node]

        return responses


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: transact.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import types_pb2 as types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0etransact.proto\x12\x10\x61\x65rospike.vector\x1a\x1bgoogle/protobuf/empty.proto\x1a\x0btypes.proto\"\xa5\x01\n\nPutRequest\x12\"\n\x03key\x18\x01 \x01(\x0b\x32\x15.aerospike.vector.Key\x12.\n\twriteType\x18\x02 \x01(\x0e\x32\x1b.aerospike.vector.WriteType\x12\'\n\x06\x66ields\x18\x03 \x03(\x0b\x32\x17.aerospike.vector.Field\x12\x1a\n\x12ignoreMemQueueFull\x18\x04 \x01(\x08\"f\n\nGetRequest\x12\"\n\x03key\x18\x01 \x01(\x0b\x32\x15.aerospike.vector.Key\x12\x34\n\nprojection\x18\x02 \x01(\x0b\x32 .aerospike.vector.ProjectionSpec\"3\n\rExistsRequest\x12\"\n\x03key\x18\x01 \x01(\x0b\x32\x15.aerospike.vector.Key\"3\n\rDeleteRequest\x12\"\n\x03key\x18\x01 \x01(\x0b\x32\x15.aerospike.vector.Key\"b\n\x10IsIndexedRequest\x12\"\n\x03key\x18\x01 \x01(\x0b\x32\x15.aerospike.vector.Key\x12*\n\x07indexId\x18\x02 \x01(\x0b\x32\x19.aerospike.vector.IndexId\"R\n\x10ProjectionFilter\x12.\n\x04type\x18\x01 \x01(\x0e\x32 .aerospike.vector.ProjectionType\x12\x0e\n\x06\x66ields\x18\x02 \x03(\t\"z\n\x0eProjectionSpec\x12\x33\n\x07include\x18\x01 \x01(\x0b\x32\".aerospike.vector.ProjectionFilter\x12\x33\n\x07\x65xclude\x18\x02 \x01(\x0b\x32\".aerospike.vector.ProjectionFilter\"\x83\x02\n\x13VectorSearchRequest\x12(\n\x05index\x18\x01 \x01(\x0b\x32\x19.aerospike.vector.IndexId\x12-\n\x0bqueryVector\x18\x02 \x01(\x0b\x32\x18.aerospike.vector.Vector\x12\r\n\x05limit\x18\x03 \x01(\r\x12\x34\n\nprojection\x18\x04 \x01(\x0b\x32 .aerospike.vector.ProjectionSpec\x12>\n\x10hnswSearchParams\x18\x05 \x01(\x0b\x32\".aerospike.vector.HnswSearchParamsH\x00\x42\x0e\n\x0csearchParams*X\n\tWriteType\x12\n\n\x06UPSERT\x10\x00\x12\x0f\n\x0bUPDATE_ONLY\x10\x01\x12\x0f\n\x0bINSERT_ONLY\x10\x02\x12\x0b\n\x07REPLACE\x10\x03\x12\x10\n\x0cREPLACE_ONLY\x10\x04*2\n\x0eProjectionType\x12\x07\n\x03\x41LL\x10\x00\x12\x08\n\x04NONE\x10\x01\x12\r\n\tSPECIFIED\x10\x02\x32\xc3\x03\n\x0fTransactService\x12=\n\x03Put\x12\x1c.aerospike.vector.PutRequest\x1a\x16.google.protobuf.Empty\"\x00\x12?\n\x03Get\x12\x1c.aerospike.vector.GetRequest\x1a\x18.aerospike.vector.Record\"\x00\x12\x43\n\x06\x44\x65lete\x12\x1f.aerospike.vector.DeleteRequest\x1a\x16.google.protobuf.Empty\"\x00\x12\x46\n\x06\x45xists\x12\x1f.aerospike.vector.ExistsRequest\x1a\x19.aerospike.vector.Boolean\"\x00\x12L\n\tIsIndexed\x12\".aerospike.vector.IsIndexedRequest\x1a\x19.aerospike.vector.Boolean\"\x00\x12U\n\x0cVectorSearch\x12%.aerospike.vector.VectorSearchRequest\x1a\x1a.aerospike.vector.Neighbor\"\x00\x30\x01\x42\x43\n!com.aerospike.vector.client.protoP\x01Z\x1c\x61\x65rospike.com/vector/protos/b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'transact_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n!com.aerospike.vector.client.protoP\001Z\034aerospike.com/vector/protos/'
  _globals['_WRITETYPE']._serialized_start=1026
  _globals['_WRITETYPE']._serialized_end=1114
  _globals['_PROJECTIONTYPE']._serialized_start=1116
  _globals['_PROJECTIONTYPE']._serialized_end=1166
  _globals['_PUTREQUEST']._serialized_start=79
  _globals['_PUTREQUEST']._serialized_end=244
  _globals['_GETREQUEST']._serialized_start=246
  _globals['_GETREQUEST']._serialized_end=348
  _globals['_EXISTSREQUEST']._serialized_start=350
  _globals['_EXISTSREQUEST']._serialized_end=401
  _globals['_DELETEREQUEST']._serialized_start=403
  _globals['_DELETEREQUEST']._serialized_end=454
  _globals['_ISINDEXEDREQUEST']._serialized_start=456
  _globals['_ISINDEXEDREQUEST']._serialized_end=554
  _globals['_PROJECTIONFILTER']._serialized_start=556
  _globals['_PROJECTIONFILTER']._serialized_end=638
  _globals['_PROJECTIONSPEC']._serialized_start=640
  _globals['_PROJECTIONSPEC']._serialized_end=762
  _globals['_VECTORSEARCHREQUEST']._serialized_start=765
  _globals['_VECTORSEARCHREQUEST']._serialized_end=1024
  _globals['_TRANSACTSERVICE']._serialized_start=1169
  _globals['_TRANSACTSERVICE']._serialized_end=1620
# @@protoc_insertion_point(module_scope)
